// fixed bits
#define IMMEDIATE		0x8000
#define WRITEBACK		0x0100
#define WITH_LINK		0x0100

// register fields
#define RB				0
#define	RA				4
#define RD				4
#define IMM				16

// argument flags
#define ARGFLAG_R4		(ARGFLAG_REG | ARGFLAG_4 | ARGFLAG_U)
#define ARGFLAG_I16		(ARGFLAG_IMM | ARGFLAG_16 | ARGFLAG_U)

// the instructions :)
const Instruction r16_instructions[] =
{
{0,	"NOP",			0,							0x18<<8, 2, 0, {}},

{0,	"ADD R~0,R~1",	0,				WRITEBACK | 0x00<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"ADD R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x00<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"ADDC R~0,R~1",	0,				WRITEBACK | 0x02<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"ADDC R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x02<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"SUB R~0,R~1",	0,				WRITEBACK | 0x04<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"SUB R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x04<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"CMP R~0,R~1",	0,							0x04<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"CMP R~0,~1",	0,	IMMEDIATE | 			0x04<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"SUBC R~0,R~1",	0,				WRITEBACK | 0x06<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"SUBC R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x06<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
                                                               
{0,	"OR R~0,R~1",	0,				WRITEBACK | 0x08<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"OR R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x08<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"AND R~0,R~1",	0,				WRITEBACK | 0x0A<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"AND R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x0A<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"TEST R~0,R~1",	0,				 			0x0A<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"TEST R~0,~1",	0,	IMMEDIATE |  			0x0A<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"XOR R~0,R~1",	0,				WRITEBACK | 0x0C<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"XOR R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x0C<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"ANDN R~0,R~1",	0,				WRITEBACK | 0x0E<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"ANDN R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x0E<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},

{0,	"SWAP R~0,R~1",	0,				WRITEBACK | 0x10<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"SWAP R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x10<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"LSL R~0,R~1",	0,				WRITEBACK | 0x12<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"LSL R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x12<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"LSR R~0,R~1",	0,				WRITEBACK | 0x14<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"LSR R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x14<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"ASR R~0,R~1",	0,				WRITEBACK | 0x16<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"ASR R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x16<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},

{0,	"MOV R~0,R~1",	0,				WRITEBACK | 0x18<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"MOV R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x18<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"SXB R~0,R~1",	0,				WRITEBACK | 0x1A<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"SXB R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x1A<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"MTC R~0,R~1",	0,				 			0x1C<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"MTC R~0,~1",	0,	IMMEDIATE |  			0x1C<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"MFC R~0,R~1",	0,				WRITEBACK | 0x1E<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"MFC R~0,~1",	0,	IMMEDIATE | WRITEBACK | 0x1E<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},

// store/load (note: absolute address)
{0,	"SB [R~0],R~1",	0,							0x20<<8, 2, 2, {{RB,ARGFLAG_R4}, {RA,ARGFLAG_R4}}},
{0,	"SB [~0],R~1",	0,	IMMEDIATE | 			0x20<<8, 4, 2, {{IMM,ARGFLAG_I16}, {RA,ARGFLAG_R4}}},
{0,	"LB R~0,[R~1]",	0,				WRITEBACK | 0x20<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"LB R~0,[~1]",	0,	IMMEDIATE | WRITEBACK | 0x20<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"SW [R~0],R~1",	0,				 			0x22<<8, 2, 2, {{RB,ARGFLAG_R4}, {RA,ARGFLAG_R4}}},
{0,	"SW [~0],R~1",	0,	IMMEDIATE |  			0x22<<8, 4, 2, {{IMM,ARGFLAG_I16}, {RA,ARGFLAG_R4}}},
{0,	"LW R~0,[R~1]",	0,				WRITEBACK | 0x22<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"LW R~0,[~1]",	0,	IMMEDIATE | WRITEBACK | 0x22<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},

// branch
{0,	"BRA R~0",		0,							0x40<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BRA ~0",		0,	IMMEDIATE | 			0x40<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BRAI R~0",		0,							0x42<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BRAI ~0",		0,	IMMEDIATE | 			0x42<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BHI R~0",		0,							0x44<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BHI ~0",		0,	IMMEDIATE | 			0x44<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BLS R~0",		0,							0x46<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BLS ~0",		0,	IMMEDIATE | 			0x46<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BLT R~0",		0,							0x48<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BLT ~0",		0,	IMMEDIATE | 			0x48<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BGE R~0",		0,							0x4A<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BGE ~0",		0,	IMMEDIATE | 			0x4A<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BLE R~0",		0,							0x4C<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BLE ~0",		0,	IMMEDIATE | 			0x4C<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BGT R~0",		0,							0x4E<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BGT ~0",		0,	IMMEDIATE | 			0x4E<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BNS R~0",		0,							0x50<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BNS ~0",		0,	IMMEDIATE | 			0x50<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BNC R~0",		0,							0x52<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BNC ~0",		0,	IMMEDIATE | 			0x52<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BVS R~0",		0,							0x54<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BVS ~0",		0,	IMMEDIATE | 			0x54<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BVC R~0",		0,							0x56<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BVC ~0",		0,	IMMEDIATE | 			0x56<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BEQ R~0",		0,							0x58<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BEQ ~0",		0,	IMMEDIATE | 			0x58<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BNE R~0",		0,							0x5A<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BNE ~0",		0,	IMMEDIATE | 			0x5A<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BCS R~0",		0,							0x5C<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BCS ~0",		0,	IMMEDIATE | 			0x5C<<8, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BCC R~0",		0,							0x5E<<8, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BCC ~0",		0,	IMMEDIATE | 			0x5E<<8, 4, 1, {{IMM,ARGFLAG_I16}}},

// return to R14
{0,	"RET",			0,							0x40<<8 | 14<<RB, 2, 0, {}},
{0,	"RETHI",		0,							0x44<<8 | 14<<RB, 2, 0, {}},
{0,	"RETLS",		0,							0x46<<8 | 14<<RB, 2, 0, {}},
{0,	"RETLT",		0,							0x48<<8 | 14<<RB, 2, 0, {}},
{0,	"RETGE",		0,							0x4A<<8 | 14<<RB, 2, 0, {}},
{0,	"RETLE",		0,							0x4C<<8 | 14<<RB, 2, 0, {}},
{0,	"RETGT",		0,							0x4E<<8 | 14<<RB, 2, 0, {}},
{0,	"RETNS",		0,							0x50<<8 | 14<<RB, 2, 0, {}},
{0,	"RETNC",		0,							0x52<<8 | 14<<RB, 2, 0, {}},
{0,	"RETVS",		0,							0x54<<8 | 14<<RB, 2, 0, {}},
{0,	"RETVC",		0,							0x56<<8 | 14<<RB, 2, 0, {}},
{0,	"RETEQ",		0,							0x58<<8 | 14<<RB, 2, 0, {}},
{0,	"RETNE",		0,							0x5A<<8 | 14<<RB, 2, 0, {}},
{0,	"RETCS",		0,							0x5C<<8 | 14<<RB, 2, 0, {}},
{0,	"RETCC",		0,							0x5E<<8 | 14<<RB, 2, 0, {}},

// return to R15
{0,	"RETI",			0,							0x42<<8 | 15<<RB, 2, 0, {}},

// branch with link
{0,	"BRAL R~0,R~1",		0,				WITH_LINK | 0x40<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BRAL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x40<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BRAIL R~0,R~1",	0,				WITH_LINK | 0x42<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BRAIL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x42<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BHIL R~0,R~1",		0,				WITH_LINK | 0x44<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BHIL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x44<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BLSL R~0,R~1",		0,				WITH_LINK | 0x46<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BLSL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x46<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BLTL R~0,R~1",		0,				WITH_LINK | 0x48<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BLTL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x48<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BGEL R~0,R~1",		0,				WITH_LINK | 0x4A<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BGEL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x4A<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BLEL R~0,R~1",		0,				WITH_LINK | 0x4C<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BLEL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x4C<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BGTL R~0,R~1",		0,				WITH_LINK | 0x4E<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BGTL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x4E<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BNSL R~0,R~1",		0,				WITH_LINK | 0x50<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BNSL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x50<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BNCL R~0,R~1",		0,				WITH_LINK | 0x52<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BNCL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x52<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BVSL R~0,R~1",		0,				WITH_LINK | 0x54<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BVSL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x54<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BVCL R~0,R~1",		0,				WITH_LINK | 0x56<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BVCL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x56<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BEQL R~0,R~1",		0,				WITH_LINK | 0x58<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BEQL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x58<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BNEL R~0,R~1",		0,				WITH_LINK | 0x5A<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BNEL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x5A<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BCSL R~0,R~1",		0,				WITH_LINK | 0x5C<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BCSL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x5C<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},
{0,	"BCCL R~0,R~1",		0,				WITH_LINK | 0x5E<<8, 2, 2, {{RD,ARGFLAG_R4}, {RB,ARGFLAG_R4}}},
{0,	"BCCL R~0,~1",		0,	IMMEDIATE | WITH_LINK | 0x5E<<8, 4, 2, {{RD,ARGFLAG_R4}, {IMM,ARGFLAG_I16}}},

// branch with link to R14
{0,	"BRAL R~0",			0,				WITH_LINK | 0x40<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BRAL ~0",			0,	IMMEDIATE | WITH_LINK | 0x40<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BRAIL R~0",		0,				WITH_LINK | 0x42<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BRAIL ~0",			0,	IMMEDIATE | WITH_LINK | 0x42<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BHIL R~0",			0,				WITH_LINK | 0x44<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BHIL ~0",			0,	IMMEDIATE | WITH_LINK | 0x44<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BLSL R~0",			0,				WITH_LINK | 0x46<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BLSL ~0",			0,	IMMEDIATE | WITH_LINK | 0x46<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BLTL R~0",			0,				WITH_LINK | 0x48<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BLTL ~0",			0,	IMMEDIATE | WITH_LINK | 0x48<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BGEL R~0",			0,				WITH_LINK | 0x4A<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BGEL ~0",			0,	IMMEDIATE | WITH_LINK | 0x4A<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BLEL R~0",			0,				WITH_LINK | 0x4C<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BLEL ~0",			0,	IMMEDIATE | WITH_LINK | 0x4C<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BGTL R~0",			0,				WITH_LINK | 0x4E<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BGTL ~0",			0,	IMMEDIATE | WITH_LINK | 0x4E<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BNSL R~0",			0,				WITH_LINK | 0x50<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BNSL ~0",			0,	IMMEDIATE | WITH_LINK | 0x50<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BNCL R~0",			0,				WITH_LINK | 0x52<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BNCL ~0",			0,	IMMEDIATE | WITH_LINK | 0x52<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BVSL R~0",			0,				WITH_LINK | 0x54<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BVSL ~0",			0,	IMMEDIATE | WITH_LINK | 0x54<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BVCL R~0",			0,				WITH_LINK | 0x56<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BVCL ~0",			0,	IMMEDIATE | WITH_LINK | 0x56<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BEQL R~0",			0,				WITH_LINK | 0x58<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BEQL ~0",			0,	IMMEDIATE | WITH_LINK | 0x58<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BNEL R~0",			0,				WITH_LINK | 0x5A<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BNEL ~0",			0,	IMMEDIATE | WITH_LINK | 0x5A<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BCSL R~0",			0,				WITH_LINK | 0x5C<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BCSL ~0",			0,	IMMEDIATE | WITH_LINK | 0x5C<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
{0,	"BCCL R~0",			0,				WITH_LINK | 0x5E<<8 | 14<<RD, 2, 1, {{RB,ARGFLAG_R4}}},
{0,	"BCCL ~0",			0,	IMMEDIATE | WITH_LINK | 0x5E<<8 | 14<<RD, 4, 1, {{IMM,ARGFLAG_I16}}},
};
